// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit TimeHelper;

interface

uses
  SysUtils, Windows, WinSock, orbtypes, Timebase_int;

const
  // Max TimeT value
  MaxTimeT = _ulonglong($FFFFFFFFFFFFFFFF);
  // Max InaccruacyT value
  MaxInaccuracyT = _ulonglong($FFFFFFFFFFFF);

type
  // This is a simpler set of functions to work with TimeBase::UtcT, and
  // TimeBase::TimeT values.
  //
  // The Time Service uses units of 100 nanoseconds and a base time
  // of 15 October 1582 00:00:00.
  TTimeHelper = class(TObject)
  public
    // Get current time in unit of 100 ns
    class function utcNow(AInaccuracy: TTimeBase_InaccuracyT = 0): TTimeBase_UtcT;
    class function utcMin(): TTimeBase_UtcT;
    class function utcMax(): TTimeBase_UtcT;
    class function toTimeval(ATimeT: TTimeBase_UtcT): timeval; overload;
    class function toTimeval(ATimeT: TTimeBase_TimeT): timeval; overload;
    class function toIntervalT(ATimeT: TTimeBase_TimeT; AInaccuracyT: TTimeBase_InaccuracyT): TTimeBase_IntervalT;
    class function toUtcT(ATimeT: TTimeBase_TimeT; AInaccuracyT: TTimeBase_InaccuracyT; ATdfT: TTimeBase_TdfT = 0): TTimeBase_UtcT; overload;
    class function toUtcT(const inter: TTimeBase_IntervalT): TTimeBase_UtcT; overload;
    class function toString(const ATime: TTimeBase_UtcT): string; overload;
    class function toTimeString(const ATime: TTimeBase_UtcT): string;
    class function toString(const ATime: TTimeBase_TimeT): string; overload;
    // Delphi support routines
    class function toDateTime(const ATime: TTimeBase_UtcT): TDateTime;
    class function toUtcT(const ADateTime: TDateTime; const AInaccuracy: TTimeBase_InaccuracyT = 0): TTimeBase_UtcT; overload;
  end;

  // operators
  function UtcT_Plus_Time(const AUtcT: TTimeBase_UtcT; const ATimeT: TTimeBase_TimeT): TTimeBase_UtcT;

implementation

uses throw;

const
  // 100 nanoseconds = 10^-7 seconds, hence there are 10^7 (10,000,000)
  // seconds per 100 nanoseconds.
  HNanosPerSecond = _ulonglong(10000000);
  // 100 ns units from 15 October 1582 00:00:00 to 1 January 1970 00:00:00.
  //
  // Time difference in 100 ns units between DCE and POSIX time base.
  // 141427 days with 86400 seconds each.
  DCEToPosix = _ulonglong(_ulonglong(141427) * _ulonglong(86400) * HNanosPerSecond);

  //DefaultTimeZone = long(5);
  //DefaultDaylight = 1;
  //_daylight = 0; //DefaultDaylight;
  _timezone = -14400; //long(DefaultTimeZone * long(60) * long(60)); // Set for EST
  _YDays: array [0..12] of cardinal = (0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365);
  _Days: array [0..11] of byte = (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
  Thursday = 4;
  April = 4;
  October = 10;

type
  time_t = long;
  timeb = record
    time: long;
    millitm,
    timezone,
    dstflag: short;
  end;
  tm = record
    tm_sec,
    tm_min,
    tm_hour,
    tm_mday,
    tm_mon,
    tm_year,
    tm_wday,
    tm_yday,
    tm_isdst: integer;
  end;
  Ptm = ^tm;

  function _daylight: integer;
  var
    tz: TIME_ZONE_INFORMATION;
  begin
    if (GetTimeZoneInformation(tz) = TIME_ZONE_ID_DAYLIGHT) then
      result := 1
    else
      result := 0;
  end;

  // determines whether daylight savings is in effect
  function _isDST(hour, yday, month, year: Cardinal): integer;
  var
    temp, sunday: Cardinal;
  begin
    if month = 0 then begin // if only day of year given
      temp := yday;
      if (yday >= 31 + 28) and (((year + 70) and 3) = 0) then
        temp := Pred(temp);
      while temp >= _YDays[month] do month := Succ(month);
    end
    else begin // if month+day of month given
      Inc(yday, _YDays[month - 1]);
      if (month > 2) and (((year + 70) and 3) = 0) then // leap year, Mar-Dec
        yday := Succ(yday);
    end;

    // Weed out months other than April or October.
    if (month < April) or (month > October) then begin
      result := 0;
      Exit;
    end;
    if (month > April) and (month < October) then begin
      result := 1;
      Exit;
    end;
    // Month is either April or October.  Up through 1986, the starting
    // day for DST is the last Sunday in April; after that it is
    // the first Sunday in April.  The last day of DST is the last Sunday
    // in October.
    if (year > 16) and (month = April) then // if April and year > 1986
      sunday := _YDays[month-1] + 7 // day = 7th day in month
    else
      sunday := _YDays[month]; // day = last day in month 
    if ((year + 70) and 3) = 1 then // leap year ?
      sunday := Pred(sunday); // no --> adjust

    // Adjust the threshold day downward to the preceding Sunday.
    temp := sunday + ((year + 1) shr 2) // add leap days since 1970
           + (year * 365)               // add years since 1970
           + Thursday;                  // 01-01-70 was Thursday 
    Dec(sunday, temp mod 7);
    
    // Check if the day falls within the limit.
    if (month = April) then begin
      if (yday > sunday) then
        result := 1
      else if (yday < sunday) then
        result := 0
      else begin
        if hour >= 2 then
          result := 1 // DST started at 2:00 AM
        else
          result := 0;
      end;
    end
    else begin
      if (yday < sunday) then
        result := 1
      else if (yday > sunday) then
        result := 0
      else begin
        if hour <= 1 then
          result := 1 // DST ended at 2:00 AM
        else
          result := 0;
      end;
    end;
  end;

  // convert a date to seconds since 1970
  function _totalSec(year, month, day, hour, min, sec, dst_flag: integer): long;
  var
    leaps: integer;
    days, secs: time_t;
  begin
    if (year < 70) or (year > 138) then begin
      result := -1;
      Exit;
    end;

    Inc(min, sec div 60);
    sec := sec mod 60; // Seconds are normalized
    Inc(hour, min div 60);
    min := min mod 60; // Minutes are normalized
    Inc(day, hour div 24);
    hour := hour mod 24; // Hours are normalized   

    Inc(year, month div 12); // Normalize month (not necessarily final)
    month := month mod 12;

    while (day >= _Days[month]) do begin
      if not ((year and 3) = 1) and (month = 1) then begin
        if (day > 28) then begin
          Dec(day, 29);
          month := Succ(month);
        end
        else
          Break;
      end
      else begin
        Dec(day, _Days[month]);
        month := Succ(month);
      end;

      if (month >= 12) then begin // Normalize month 
        Dec(month, 12);
        year := Succ(year);
      end
    end;

    Dec(year, 70);
    leaps := (year + 2) div 4;

    if not(((year+70) and 3) = 1) and (month < 2) then
      leaps := Pred(leaps);

    days := year * 365 + leaps + integer(_YDays[month]) + day;

    secs := days * 86400 + hour * 3600 + min * 60 + sec + _timezone;

    if (dst_flag = -1) then begin
      if (_daylight = 1) and (_isDST(hour, day, month+1, year) = 1) then
        Dec(secs, 3600);
    end
    else begin
      if dst_flag = 1 then
        Dec(secs, 3600);
    end;

    if secs > 0 then
      result := secs
    else
      result := -1;
  end;

  // Get current time and store it in 'timeb' structure
  procedure ftime(var timeStruct: timeb);
  var
    d: SYSTEMTIME;
  begin
    GetLocalTime(d);
    // Convert extern 'timezone's seconds to structure 'timezone's minutes.
    // Set daylight savings indicator.
    // Convert NT date and time to UNIX style time and store in structure.
    with timeStruct, d do begin
      timezone := short(_timezone div long(60));
      dstflag := _daylight and _isDST(wHour, wDay, wMonth, wYear - 1970);
      time := long(_totalsec(wYear - 1900, wMonth-1, wDay-1, wHour, wMinute, wSecond, -1));
      millitm := wMilliseconds;
    end; { with }
  end;

  //fills the time structure tm by translating the long time.
  function comtime(ltime: time_t; dst: integer): tm;
  var
    _time: ulong;
    hpery: integer;
    i, cumdays: cardinal;
  begin
    _time := ulong(ltime);
    result.tm_sec := integer(_time mod 60);
    _time := _time div 60; // Time in minutes 
    result.tm_min := integer(_time mod 60);
    _time := _time div 60; // Time in hours 
    i := cardinal(_time div (1461 * 24)); // Number of 4 year blocks
    result.tm_year := (i shl 2);
    Inc(result.tm_year, 70);
    cumdays := 1461 * i;
    _time := _time mod (1461 * 24); // Hours since end of last 4 year block 

    while true do begin
        hpery := 365 * 24;
        if (result.tm_year mod 3) = 0 then
          Inc(hpery, 24);
        if _time < ulong(hpery) then
          Break;
        Inc(cumdays, hpery div 24);
        result.tm_year := Succ(result.tm_year);
        Dec(_time, hpery);
    end; // at end, time is number of hours into current year

    if  ((dst and _daylight) = 1) and (_isDST(_time mod 24, _time div 24, 0, result.tm_year - 70) = 1) then begin
        _time := Succ(_time);
        result.tm_isdst := 1;
    end
    else
      result.tm_isdst := 0;

    result.tm_hour := _time mod 24;
    _time := _time div 24; // Time in days 
    result.tm_yday := _time;
    Inc(cumdays, _time + 4);
    result.tm_wday := cumdays mod 7;
    _time := Succ(_time);

    if ((result.tm_year and 3) = 0) then begin
        if (_time > 60) then
            _time := Pred(_time)
        else
          if (_time = 60) then begin
            result.tm_mon := 1;
            result.tm_mday := 29;
            Exit;
          end;
    end;

    result.tm_mon := 0;
    while _Days[result.tm_mon] < _time do begin
      Dec(_time, _Days[result.tm_mon]);
      result.tm_mon := Succ(result.tm_mon);
    end;

    result.tm_mday := _time;
  end;
  // converts date and time to Greenwich Mean Time
  function gmTime(const clock: time_t): tm;
  begin
    result := comtime(clock, 0);
  end;

// operators
function UtcT_Plus_Time(const AUtcT: TTimeBase_UtcT; const ATimeT: TTimeBase_TimeT): TTimeBase_UtcT;
begin
  result := AUtcT;
  if MaxTimeT - result.time < ATimeT then
    result.time := MaxTimeT
  else
    Inc(result.time, ATimeT);
end;

{ TTimeHelper }

class function TTimeHelper.toUtcT(
  const ADateTime: TDateTime; const AInaccuracy: TTimeBase_InaccuracyT): TTimeBase_UtcT;
var
  st: SYSTEMTIME;
  timebuffer: timeb;
  _time: TTimeBase_TimeT;
  tz: TIME_ZONE_INFORMATION;
  tdf: TTimeBase_TdfT;
begin
  DateTimeToSystemTime(ADateTime, st);
{$IFDEF WIN32}
  // Convert extern 'timezone's seconds to structure 'timezone's minutes.
  // Set daylight savings indicator.
  // Convert NT date and time to UNIX style time and store in structure.
  with timebuffer, st do begin
    timezone := short(_timezone div long(60));
    dstflag := _daylight and _isDST(wHour, wDay, wMonth, wYear - 1970);
    time := long(_totalsec(wYear - 1900, wMonth-1, wDay-1, wHour, wMinute, wSecond, -1));
    millitm := wMilliseconds;
  end; { with }
  _time := _ulonglong(timebuffer.time) * HNanosPerSecond + _ulonglong(timebuffer.millitm) * 10000 + DCEToPosix;
  if (GetTimeZoneInformation(tz) = TIME_ZONE_ID_DAYLIGHT) then
    tdf := -(Short(tz.Bias) + Short(tz.DaylightBias))
  else
    tdf := -(Short(tz.Bias));
{$ENDIF}
  result := toUtcT(_time, AInaccuracy, tdf);
end;

class function TTimeHelper.toDateTime(
  const ATime: TTimeBase_UtcT): TDateTime;
var
  tv: timeval;
  lt: tm;
begin
  // Get local time
  tv := toTimeval(ATime.time);
  Inc(tv.tv_sec, ATime.tdf * 60);
  lt := gmtime(tv.tv_sec);

  // Convert to string
  with lt do
    result := EncodeDate(1900 + tm_year, tm_mon + 1, tm_mday) + EncodeTime(tm_hour, tm_min, tm_sec, tv.tv_usec div 1000);
end;

class function TTimeHelper.toIntervalT(ATimeT: TTimeBase_TimeT;
  AInaccuracyT: TTimeBase_InaccuracyT): TTimeBase_IntervalT;
begin
  with result do begin
    if ATimeT < AInaccuracyT then
      lower_bound := 0
    else
      lower_bound := ATimeT - AInaccuracyT;

    if MaxTimeT - ATimeT < AInaccuracyT then
      upper_bound := MaxTimeT
    else
      upper_bound := ATimeT - AInaccuracyT;
  end; { with }
end;

class function TTimeHelper.toString(const ATime: TTimeBase_UtcT): string;
begin
  if ATime.time < DCEToPosix then begin
    result := 'Time less than 1 January 1970 00:00:00';
    Exit;
  end;

  // Convert to string
  result := DateTimeToStr(toDateTime(ATime));
end;

class function TTimeHelper.toString(const ATime: TTimeBase_TimeT): string;
var
  milliTime: TTimeBase_TimeT;
  sec, msec: _ulong;
begin
  milliTime := ATime div 10000;

  sec := milliTime div 1000;
  msec := milliTime mod 1000;

  result := IntToStr(sec) + ':';
  if (msec < 10) then
      result := result + '00'
  else if (msec < 100) then
    result := result + '0';
  result := result + IntToStr(msec);
end;

class function TTimeHelper.toTimeString(
  const ATime: TTimeBase_UtcT): string;
begin
  if ATime.time < DCEToPosix then begin
    result := 'Time less than 1 January 1970 00:00:00';
    Exit;
  end;

  result := TimeToStr(toDateTime(ATime));
end;

class function TTimeHelper.toTimeval(ATimeT: TTimeBase_TimeT): timeval;
var
  v: _ulonglong;
begin
  if ATimeT < DCEToPosix then begin
    result.tv_sec := 0;
    result.tv_usec := 0;
    Exit;
  end;

  v := ATimeT - DCEToPosix;

  result.tv_sec := v div HNanosPerSecond;
  result.tv_usec := (v mod HNanosPerSecond) div 10;
end;

class function TTimeHelper.toTimeval(ATimeT: TTimeBase_UtcT): timeval;
begin
  result := toTimeVal(ATimeT.time);
end;

class function TTimeHelper.toUtcT(ATimeT: TTimeBase_TimeT;
  AInaccuracyT: TTimeBase_InaccuracyT;
  ATdfT: TTimeBase_TdfT): TTimeBase_UtcT;
begin
  with result do begin
    time := ATimeT;
    inacclo := long(AInaccuracyT and $FFFFFFFF);
    inacchi := short((AInaccuracyT shr 32) and $FFFF);
    tdf := ATdfT;
  end; { with }
end;

class function TTimeHelper.toUtcT(
  const inter: TTimeBase_IntervalT): TTimeBase_UtcT;
var
  inaccuracy: TTimeBase_InaccuracyT;
  _time: TTimeBase_TimeT;
begin
  inaccuracy := (inter.upper_bound - inter.lower_bound) div 2;
  _time := inter.lower_bound + inaccuracy;

  result := toUtcT(_time, inaccuracy);
end;

class function TTimeHelper.utcMax: TTimeBase_UtcT;
begin
  result := toUtcT(MaxTimeT, 0, utcNow(0).tdf);
end;

class function TTimeHelper.utcMin: TTimeBase_UtcT;
begin
  result := toUtcT(0, 0, utcNow(0).tdf);
end;

class function TTimeHelper.utcNow(
  AInaccuracy: TTimeBase_InaccuracyT): TTimeBase_UtcT;
var
  timebuffer: timeb;
  _time: TTimeBase_TimeT;
  tz: TIME_ZONE_INFORMATION;
  tdf: TTimeBase_TdfT;
begin
  if AInaccuracy > MaxInaccuracyT then
    dorb_throw(st_BAD_PARAM);
{$IFDEF WIN32}
  ftime(timebuffer);
  _time := _ulonglong(timebuffer.time) * HNanosPerSecond + _ulonglong(timebuffer.millitm) * 10000 + DCEToPosix;
  if (GetTimeZoneInformation(tz) = TIME_ZONE_ID_DAYLIGHT) then
    tdf := -(Short(tz.Bias) + Short(tz.DaylightBias))
  else
    tdf := -(Short(tz.Bias));
{$ENDIF}
  result := toUtcT(_time, AInaccuracy, tdf);
end;

end.
