// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit proxy;

interface

uses
  orb_int,orb,Classes,orbtypes,SysUtils,giop_int,addr_int,mtmanager,mtmanager_int,
  code_int,comp_int,iiop_int,disp_int,req_int,map_int,env_int,iop,osthread;

{$I dorb.inc}

{$IFNDEF USELIB}

type

  TIIOPProxy = class(TORBObject,IObjectAdaptor,IGIOPCallBack)
  private
    FConnections : TStringList;
    FConnMutex: TMutex;
    FInvMutex: TMutex;
{$IFDEF USE_IOP_CACHE}
    FCacheUsed : Boolean;
    FCacheRecord : IIIOPProxyInvokeRecord;
{$ENDIF}
    FInvokes : ILongInterfaceMap;
    FConnMgr: TGIOPConnectionMgr;
    FGIOPVersion: _ushort;
    procedure deref_connection(const conn: IGIOPConnection; all: boolean = false);
  protected
    function waitfor(const orb: IORB; id: ORBMessageID; event: TORBCallbackEvent; tmout: Longint = -1): boolean;
    procedure notify(const orb: IORB; id: ORBMessageID; event: TORBCallbackEvent);
    function input_giop_callback(const conn: IGIOPConnection;const buf: IBuffer{;event: TEventType}): Boolean;
    procedure sendORBMsg(const conn: IGIOPConnection;event: TORBMsgEvent);
    procedure connection_error(const conn: IGIOPConnection; senderror : Boolean);
    procedure callback(const orb: IORB; mid: MessageID; event: TORBCallbackEvent);
    procedure answer_invoke(mid: ORBMessageID; const obj: IORBObject; const req: IORBRequest; stat: TInvokeStatus);
    function skeleton(const obj: IORBObject): IORBObject;
    function kill_connection(const conn: IGIOPConnection; redo : Boolean): Boolean;
    function make_connection(const addr: IAddress): IGIOPConnection;
    function invoke(mid: ORBMessageID;const obj: IORBObject;const req: IORBRequest; const pr: IPrincipal; response_exp: Boolean): Boolean; stdcall;
    function has_object(const obj: IORBObject): Boolean;
    function get_codeset_ids(const obj: IORBObject;var csid,wcsid: CodesetID): Boolean;
    function handle_invoke_reply(const conn: IGIOPConnection;const incon: IGIOPInContext): Boolean;
    function handle_input(const conn: IGIOPConnection; const inp: IBuffer): Boolean;
    function giop_callback(const conn: IGIOPConnection;event: TEventType): Boolean;
    function create_invoke(): IIIOPProxyInvokeRecord;
    procedure del_invoke(rec: IIIOPProxyInvokeRecord);
    procedure add_invoke(inv: IIIOPProxyInvokeRecord);
    function pull_invoke(mid: ORBMessageID): IIIOPProxyInvokeRecord;
    procedure exec_invoke_reply(ctx: IGIOPInContext;reqid : ORBMessageID; rtype: TReplyStatusType;
      obj: IORBObject; req: IORBRequest; conn: IGIOPConnection);
    function bind(mid: ORBMessageID; const repoid: string; const oid: IObjectTag; const addr: IAddress): Boolean;
    function get_oaid(): ObjectID;
    function is_local(): Boolean;
    procedure abort_invoke(mid: ORBMessageID);
    procedure shutdown(WaitForCompletion: Boolean);
    procedure cancel(id: ORBMessageID);
  public
    constructor Create(orb : IORB; giop_version: _ushort = $0100);
    destructor Destroy(); override;
  end;

implementation

uses
  decoder,encoder,buffer,codeset_ref,iior_int,mcomp_int,codec,
  ior,static,req,exceptions,except_int,map,utils;


{********************************************************************
  TIIOPProxy
********************************************************************}
constructor TIIOPProxy.Create(orb : IORB; giop_version: _ushort);
var
  value: string;
begin
  FOrb := orb;
  FGIOPVersion := giop_version;
  FConnections := TStringList.Create;
  FConnMutex := TMutex.Create;
  FInvMutex := TMutex.Create;
  orb.register_oa(self,OAPriorityRemote);
  FConnMgr := TGIOPConnectionMgr.Create;
{$IFDEF USE_IOP_CACHE}
  FCacheUsed := False;
  FCacheRecord := TIIOPProxyInvokeRecord.Create;
{$ENDIF}                                                   
  value := FOrb.get_properties().get_Property('mtdorb.orb.connections_limit');
  if value <> '' then
    FConnMgr.set_conn_limit(StrToInt(value))
  else
    FConnMgr.set_conn_limit(10);
  FInvokes := TLongInterfaceMap.Create;
end;

destructor TIIOPProxy.Destroy;
var
  i : integer;
begin
  FOrb.unregister_oa(Self);
  FConnMutex.lock();
  try
    for i := 0 to Pred(FConnections.Count) do
      IGIOPConnection(Pointer(FConnections.Objects[i]))._release;
  finally
    FConnMutex.unlock();
  end; { try/finally }
  FInvMutex.lock();
  try
    FInvokes := nil;
  finally
    FInvMutex.unlock();
  end; { try/finally }
  FConnections.Free;
  FConnMutex.Free;
  FInvMutex.Free;
  FInvokes := nil;
  FConnMgr.Free;
  inherited;
end;

procedure TIIOPProxy.callback(const orb: IORB; mid: MessageID; event: TORBCallbackEvent);
begin
  assert(false,'not implemented');
end;

procedure TIIOPProxy.answer_invoke(mid: ORBMessageID; const obj: IORBObject; const req: IORBRequest; stat: TInvokeStatus);
begin
  assert(false,'not implemented');
end;

function TIIOPProxy.is_local(): Boolean;
begin
  result := false;
end;

function TIIOPProxy.skeleton(const obj: IORBObject): IORBObject;
begin
  result := nil;
end;

function TIIOPProxy.get_oaid(): ObjectID;
begin
  result :=  'dorb-iiop-proxy';
end;

function TIIOPProxy.pull_invoke(mid: ORBMessageID): IIIOPProxyInvokeRecord;
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FInvMutex);
  result := nil;
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed and (FCacheRecord.id = mid) then
    begin
      result := FCacheRecord;
      exit;
    end;
{$ENDIF}
  //result := IIIOPProxyInvokeRecord(FInvokes[mid]);
  result := IIIOPProxyInvokeRecord(FORB.get_InvokeHint(mid));
  //result._Release;
  if (result <> nil) and result.active() then begin
    result.deactivate();
  end;
end;

procedure TIIOPProxy.add_invoke(inv: IIIOPProxyInvokeRecord);
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FInvMutex);
{$IFDEF USE_IOP_CACHE}
  if inv = FCacheRecord then Exit;
{$ENDIF}
  Assert(inv.active());
  FInvokes[FOrb.get_msgid(inv.id)] := inv;
  inv._AddRef;
  FOrb.set_InvokeHint(inv.id, Pointer(inv));
end;

procedure TIIOPProxy.del_invoke(rec: IIIOPProxyInvokeRecord);
var
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FInvMutex);
  Assert(not rec.active());
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed and (FCacheRecord.id = id) then
    begin
      FCacheRecord.reset;
      FCacheUsed := false;
      exit;
    end;
{$ENDIF}
  FInvokes.erase(rec.msgid());
  rec._Release;
end;

function TIIOPProxy.create_invoke(): IIIOPProxyInvokeRecord;
begin
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed then
    result := TIIOPProxyInvokeRecord.Create
  else
    begin
      FCacheUsed := True;
      result := FCacheRecord;
    end;
{$ELSE}
  result := TIIOPProxyInvokeRecord.Create;
{$ENDIF}
end;

function TIIOPProxy.make_connection(const addr: IAddress): IGIOPConnection;
var
  i : integer;
  transport : ITransport;
  sync: ISynchronized;
begin
  //FOrb.register_oa(Self,OAPriorityRemote);
  sync := TSynchronized.Create(FConnMutex);
  result := nil;
  if addr = nil then exit;
  i := FConnections.IndexOf(addr.toString);
  if i >= 0 then begin
    result := IGIOPConnection(Pointer(FConnections.Objects[i]));
    if not result.activeRef() then
      result := nil;
  end
  else
    begin
      if not FConnMgr.aquire_conn() then Exit;
      transport := addr.make_transport();
      if transport.connect(addr) then
        begin
          result := TGIOPConnection.Create(FConnMgr.Dispatcher(),FORB,transport,self,TGIOPCodec.Create(TDecoder.Create(nil),TEncoder.Create(nil),FGIOPVersion));
          result._AddRef;
          FConnections.AddObject(transport.peer.toString,Pointer(result));
          result.start();
          if not result.activeRef() then
            result := nil;
        end
      else
        FConnMgr.release_conn();
    end;
end;

procedure TIIOPProxy.exec_invoke_reply(ctx: IGIOPInContext;reqid : ORBMessageID; rtype: TReplyStatusType;
      obj: IORBObject; req: IORBRequest; conn: IGIOPConnection);
var
  orb_stat : TInvokeStatus;
  bindobj : IORBObject;
  stat : TLocateStatus;
begin
  case orb.request_type(reqid) of
    RequestInvoke:  begin
        if rtype = orbtypes.USER_EXCEPTION then
          orb_stat := orbtypes.InvokeUserEx
        else if rtype = orbtypes.SYSTEM_EXCEPTION then
          orb_stat := orbtypes.InvokeSysEx
        else if rtype = orbtypes.LOCATION_FORWARD then
          orb_stat := orbtypes.InvokeForward
        else if rtype = orbtypes.LOCATION_FORWARD_PERM then
          orb_stat := orbtypes.InvokeForward
        else
          orb_stat := orbtypes.InvokeOk;
        orb.answer_invoke(reqid,orb_stat,obj,req);
      end;
    RequestBind:  begin
        if (rtype <> orbtypes.NO_EXCEPTION) or
           (not conn.codec.get_bind_reply(ctx,stat,bindobj)) or
           (stat <> orbtypes.LocateHere) then
          orb.answer_bind(reqid,orbtypes.LocateUnknown,nil)
        else
          orb.answer_bind(reqid,orbtypes.LocateHere,bindobj)
      end;
  end;
end;

function TIIOPProxy.handle_invoke_reply(const conn: IGIOPConnection;const incon: IGIOPInContext): Boolean;
var
  reqid : _ulong;
  id: ORBMessageID;
  status : TReplyStatusType;
  rec : IIIOPProxyInvokeRecord;
  obj : IORBObject;
  ex: ISystemException;
  cd: IGIOPCodec;
  ctx: ServiceContextList;
begin
  result := false;
  if not conn.codec.get_invoke_reply1(incon,reqid,status,ctx) then
    begin
      connection_error(conn,true);
      exit;
    end;
  // turn the MsgId into somthing usefull
  id := FOrb.get_orbid(reqid);
  rec := pull_invoke(id);
  conn.activeDeref();
  if rec = nil then 	// has been canceled;
    begin
      result := true;
      exit;
    end;

  incon.set_converter(conn.codec.converter());

  cd := conn.codec as IGIOPCodec;
  if not cd.get_invoke_reply2(incon,reqid,status,obj,rec.request,ctx) then
    begin
      if rec.request <> nil then
        begin
          ex := SystemException.Create_val(0,COMPLETED_MAYBE);
          rec.request.set_out_args(ex);
          status := SYSTEM_EXCEPTION;
        end;
    end;
  exec_invoke_reply(incon,id,status,obj,rec.request,conn);
  
  del_invoke(rec);
  result := true;
end;

function TIIOPProxy.handle_input(const conn: IGIOPConnection; const inp: IBuffer): Boolean;
var
  ctx : IGIOPInContext;
  mt : TMessageType;
  size: _ulong;
  flags: byte;
begin
  result := false;
  ctx := TGIOPInContext.Create(conn.codec,inp{conn.input},nil,nil);
  if not conn.codec.get_header(ctx,mt,size,flags) then begin
    conn.activeDeref();
    exit;
  end;
  case mt of
    orbtypes.Reply: result := handle_invoke_reply(conn,ctx);
    orbtypes.LocateReply: assert(false,'not implemented');
    orbtypes.CloseConnection: begin
        if not conn.codec.get_close_msg(ctx) then begin
          conn.activeDeref();
          connection_error(conn,true);
        end
        else begin
          conn.activeDeref();
          kill_connection(conn,true); // restore
        end;
      end;
    orbtypes.MessageError : begin
        if not conn.codec.get_close_msg(ctx) then begin
          conn.activeDeref();
          connection_error(conn,false);
        end
        else begin
          conn.activeDeref();
          kill_connection(conn,false);
        end;
      end;
    else begin
      conn.activeDeref();
      assert(false,'not implemented');
    end;
  end;
end;

function TIIOPProxy.kill_connection(const conn: IGIOPConnection; redo: Boolean): Boolean;
var
  i : integer;
  found: boolean;
begin
  result := false; found := false;
  //FOrb.unregister_oa(Self);
  FConnMutex.lock();
  try
    i := FConnections.IndexOfObject(Pointer(conn));
    while i >= 0 do
      begin
        found := true;
        FConnections.Delete(i);
        conn._Release;
        i := FConnections.IndexOfObject(Pointer(conn));
      end;
  finally
    FConnMutex.unlock();
  end; { try/finally }

  if not found then Exit;

  conn.terminate();

{$IFDEF USE_IOP_CACHE}
  if FCacheUsed and (FCacheRecord.connection = conn) then
    begin
      if redo then
        assert(false,'not imlplemented')
      else
        abort_invoke(FCacheRecord.id);
    end;
{$ENDIF}
  FInvMutex.lock();
  try
    for i := Pred(FInvokes.Count) downto 0 do
      with IIIOPProxyInvokeRecord(FInvokes.Items(i)) do
        begin
          if connection = conn then
            abort_invoke(id);
          if active() then deactivate();
        end;
  finally
    FInvMutex.unlock();
  end; { try/finally }
  deref_connection(conn);
end;

function TIIOPProxy.giop_callback(const conn: IGIOPConnection;event: TEventType): Boolean;
begin
  result := true;
  case event of
    InputReady: result:= handle_input(conn, conn.input);
    Closed,Idle : begin
      kill_connection(conn,false);
      result := true;
    end;
    else
      assert(false,'not implemented');
  end;
end;

function TIIOPProxy.get_codeset_ids(const obj: IORBObject;var csid,wcsid: CodesetID): Boolean;
var
 csc : ICodesetComponent;
 prof : IIORProfile;
 mc : IMultiComponent;
 comp : IORBComponent;
begin
  prof := obj.IOR.get_profile(TAG_MULTIPLE_COMPONENTS);
  if prof <> nil then
    begin
      mc := IMultiComponentProfile(prof).get_components();
      comp := ICodesetComponent(mc.get_component(TAG_CODE_SETS));
      if comp <> nil then
        csc := ICodesetComponent(comp);
    end;
  if csc <> nil then
    begin
      csid := csc.selected_cs;
      wcsid := csc.selected_wcs;
    end
  else
    begin
      csid := $00010001;
      wcsid := $00010109;
    end;
  result := true;
end;

function TIIOPProxy.bind(mid: ORBMessageID; const repoid: string; const oid: IObjectTag; const addr: IAddress): Boolean;
var
  conn: IGIOPConnection;
  outc : IGIOPOutContext;
  rec: IIIOPProxyInvokeRecord;
begin
  result := false;
  if (addr = nil) then exit;
  conn := make_connection(addr);
  if conn = nil then
    begin
      orb.answer_bind(mid,LocateUnknown,nil);
      result := true;
      exit;
    end;
  outc := TGIOPOutContext.Create(conn.codec,nil,nil);
  conn.codec.put_bind_request(outc,FOrb.get_msgid(mid),repoid,oid);
  rec := create_invoke();
  rec.init(mid,conn,nil);
  add_invoke(rec);
  conn.output(outc.buffer);
  conn.activeDeref();
  result := true;
end;

function TIIOPProxy.invoke(mid: ORBMessageID;const obj: IORBObject;const req: IORBRequest; const pr: IPrincipal; response_exp: Boolean): Boolean;
var
  connection : IGIOPConnection;
  prof : IIORProfile;
  //csid,wcsid: CodesetID;
  //conv,wconv : ICodesetConverter;
  ocnt : IGIOPOutContext;
  rec: IIIOPProxyInvokeRecord;
  i: integer;
begin
  result := false;
  { TODO -oOVS -ctemporarily : Implement over TransportPrefPolicy }
  i := 0;
  while (i <= High(VALID_PROFILES)) and (prof = nil) do begin
    prof := obj.IOR.get_profile(VALID_PROFILES[i]);
    i := Succ(i);
  end;
  //prof := obj.IOR.get_profile(TAG_INTERNET_IOP);
  if prof <> nil then
    begin
      connection := make_connection(prof.address);
      if connection <> nil then
        begin
          {csid := 0;
          wcsid := 0;
          conv := nil;
          wconv := nil;}
          //get_codeset_ids(obj,csid,wcsid);
          if connection.codec.converter = nil then
            connection.codec.setup_codeset_ids(obj);
          ocnt := TGIOPOutContext.Create(connection.codec,connection.codec.converter,connection.codec.converter);
          if connection.codec.put_invoke_request(ocnt,FOrb.get_current_msgid(){FOrb.get_msgid(mid)},response_exp,obj,req,pr) then
            begin
              if response_exp then
                begin
                  rec := create_invoke();
                  rec.init(mid,connection,req);
                  add_invoke(rec);
                end;
              connection.output(ocnt.buffer);
              connection.activeDeref();
              result := true;
{$IFNDEF HAVE_THREADS}
              if response_exp and FConnMgr.dispatcher().isblocking() then
                connection.do_read(FConnMgr.dispatcher().isblocking());
{$ENDIF}
            end
          else
            begin
              req.set_out_args(SystemException.Create_type(st_MARSHAL) as IORBException);
              orb.answer_invoke(mid,InvokeSysEx,nil,req);
              connection.activeDeref();
            end;
        end
      else
        begin
          req.set_out_args(SystemException.Create_type(st_COMM_FAILURE) as IORBException);
          orb.answer_invoke(mid,InvokeSysEx,nil,req);
        end;
    end;
end;

function TIIOPProxy.has_object(const obj: IORBObject): Boolean;
var
  temp: IIOR;
  p1, p2: IIORProfile;
  i: integer;
begin
  result := false;
  temp := FOrb.ior_template();
  for i := 0 to High(VALID_PROFILES) do begin
    p1 := obj.IOR.get_profile(VALID_PROFILES[i]);
    if p1 <> nil then begin
      p2 := temp.get_profile(VALID_PROFILES[i]);
      result := (p2 = nil) or (not (p1 = p2));
      Break;
    end;
  end;
end;

{**  send ansower with exception }
procedure TIIOPProxy.abort_invoke(mid: ORBMessageID);
var
  obj : IORBObject;
  req : IRequest;
  local : IORBRequest;
begin
  case FORB.request_type(mid) of
    RequestInvoke: begin
        obj := TORBObject.Create(TIOR.Create as IIOR);
        req := TRequest.Create(obj,'someop');
        local := TLocalRequest.Create_req(req);
        local.set_out_args(SystemException.Create_val(0,COMPLETED_MAYBE) as IORBException);
        FORB.answer_invoke(mid,InvokeSysEx,nil,local);
      end;
    RequestBind:
        FORB.answer_bind(mid,LocateUnknown,nil);
    else
      assert(false,'not imlplemented');
  end;
  //del_invoke(mid);
end;

procedure TIIOPProxy.shutdown(WaitForCompletion: Boolean);
var
  i : integer;
  conn: IGIOPConnection;
begin
  FConnMutex.lock();
  try
    i := 0;
    while i < FConnections.Count do begin
      conn := IGIOPConnection(Pointer(FConnections.Objects[i]));
      kill_connection(conn, false);
    end; { while }
  finally
    FConnMutex.unlock();
  end; { try/finally }
  FORB.answer_shutdown(self);
end;

procedure TIIOPProxy.cancel(id: ORBMessageID);
var
  rec: IIIOPProxyInvokeRecord;
  outc : IGIOPOutContext;
begin
  rec := pull_invoke(id);
  if rec = nil then exit;
  outc := TGIOPOutContext.Create(rec.connection().codec,nil,nil);
  rec.connection.codec.put_cancel_request(outc,ORB_Instance.get_msgid(id));
  rec.connection.output(outc.buffer);
  deref_connection(rec.connection);
  del_invoke(rec);
end;

procedure TIIOPProxy.connection_error(const conn: IGIOPConnection;
  senderror: Boolean);
var
  outc : IGIOPOutContext;
begin
  if not senderror then
    kill_connection(conn,false)
  else begin
    outc := TGIOPOutContext.Create(conn.codec,nil,nil);
    conn.codec.put_error_msg(outc);
    conn.output(outc.buffer);
  end;
  //conn._Release;
  deref_connection(conn, true);
end;


function TIIOPProxy.input_giop_callback(const conn: IGIOPConnection;
  const buf: IBuffer{; event: TEventType}): Boolean;
begin
  result := handle_input(conn, buf);
end;

procedure TIIOPProxy.notify(const orb: IORB; id: ORBMessageID;
  event: TORBCallbackEvent);
begin

end;

procedure TIIOPProxy.sendORBMsg(const conn: IGIOPConnection;
  event: TORBMsgEvent);
var
  msg: IMsgType;
  orbmsg: IOrbMsg;
begin
  orbmsg := TORBMsg.Create(conn, nil, event);
  orbmsg._AddRef();
  msg := TMsgType.Create(Pointer(orbmsg));
  TPM.put_msg(ord(odORB), msg);
end;

function TIIOPProxy.waitfor(const orb: IORB; id: ORBMessageID;
  event: TORBCallbackEvent; tmout: Integer): boolean;
begin
  result := false;
end;

procedure TIIOPProxy.deref_connection(const conn: IGIOPConnection;
  all: boolean);
begin
  if conn.deref() then begin
    sendORBMsg(conn, omeKillConn);
    FConnMgr.release_conn();
  end;
end;

{$ELSE}
implementation
{$ENDIF}
end.
