// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit utils;

interface

uses
  SysUtils,orbtypes,codeset_ref,Classes;

{$IFNDEF USELIB}

  function url_encode(val:  String): string;
  function find_codeset(id:CodesetID): PCodesetInfo;
  procedure split(const value: PChar; dest: TStrings);
  function octet_to_xdigit(o: Octet): Octet;
  function byte_from_xdigit(o: Octet): Byte;
  procedure swap2(const src; var dst);
  procedure swap4(const src; var dst);
  procedure swap8(const src; var dst);
  procedure swap16(const src; var dst);
  function string_hash(str: string; max: _ulong): _ulong;

implementation

const
  digits = '0123456789abcdef';
  values : array[1..16] of byte = (0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15);

function byte_from_xdigit(o: Octet): Byte;
begin
  result := values[pos(Lowercase(o),digits)]
end;

function octet_to_xdigit(o: Octet): Octet;
begin
  result := digits[Byte(o)+1];
end;

procedure split(const value: PChar; dest: TStrings);
var
  pos,start : integer;
begin
  pos := 0;
  start := 0;
  if (value <> nil) and (dest <> nil) then
    while BYTE(value[pos]) <> 0  do
      begin
        if value[pos] = ' ' then
          begin
            if pos > start then
              dest.Add(copy(value,start+1,pos-start));
            start := pos+1;
          end;
        inc(pos);
      end;
  if pos > start then
    dest.Add(copy(value,start+1,pos-start));
end;

function find_codeset(id:CodesetID): PCodesetInfo;
var
  i : integer;
begin
  result := nil;
  for i := 0 to High(TCodesets) do
    if TCodesets[i].id = id then
      begin
        result := @TCodesets[i];
        break;
      end;
end;

function url_encode(val:  String): string;
var
  i : integer;
begin
  result := '';
  for i := 1 to Length(val) do
    case val[i] of
      'a'..'z','A'..'Z','0'..'9': result := result + val[i];
      ';','/',':','?','@','&','=','+','$',',','-','.','!','~','*','\','(',')': result := result + val[i];
      else
          result := result + '%'+ intToHex(Ord(val[i]) shl 4,1) + intToHex(Ord(val[i]) and $F,1);
    end;
end;

procedure swap2(const src; var dst);
begin
  TByteArray(dst)[0] := TByteArray(src)[1];
  TByteArray(dst)[1] := TByteArray(src)[0];
end;

procedure swap4(const src; var dst);
begin
  TByteArray(dst)[0] := TByteArray(src)[3];
  TByteArray(dst)[1] := TByteArray(src)[2];
  TByteArray(dst)[2] := TByteArray(src)[1];
  TByteArray(dst)[3] := TByteArray(src)[0];
end;

procedure swap8(const src; var dst);
begin
  TByteArray(dst)[0] := TByteArray(src)[7];
  TByteArray(dst)[1] := TByteArray(src)[6];
  TByteArray(dst)[2] := TByteArray(src)[5];
  TByteArray(dst)[3] := TByteArray(src)[4];
  TByteArray(dst)[4] := TByteArray(src)[3];
  TByteArray(dst)[5] := TByteArray(src)[2];
  TByteArray(dst)[6] := TByteArray(src)[1];
  TByteArray(dst)[7] := TByteArray(src)[0];
end;

procedure swap16(const src; var dst);
begin
  TByteArray(dst)[0] := TByteArray(src)[15];
  TByteArray(dst)[1] := TByteArray(src)[14];
  TByteArray(dst)[2] := TByteArray(src)[13];
  TByteArray(dst)[3] := TByteArray(src)[12];
  TByteArray(dst)[4] := TByteArray(src)[11];
  TByteArray(dst)[5] := TByteArray(src)[10];
  TByteArray(dst)[6] := TByteArray(src)[9];
  TByteArray(dst)[7] := TByteArray(src)[8];
  TByteArray(dst)[8] := TByteArray(src)[7];
  TByteArray(dst)[9] := TByteArray(src)[6];
  TByteArray(dst)[10] := TByteArray(src)[5];
  TByteArray(dst)[11] := TByteArray(src)[4];
  TByteArray(dst)[12] := TByteArray(src)[3];
  TByteArray(dst)[13] := TByteArray(src)[2];
  TByteArray(dst)[14] := TByteArray(src)[1];
  TByteArray(dst)[15] := TByteArray(src)[0];
end;

function string_hash(str: string; max: _ulong): _ulong;
var
  i,len : integer;
begin
  result := 0;
  if max = 0 then exit;
  len := Length(str);
  i := 1;
  if len < 16 then
    while i <= len do
      begin
        result := result * 37 + Ord(str[i]);
        inc(i);
      end
  else
    while i <= len do
      begin
        result := result * 39 + Ord(str[i]);
        inc(i,8);
      end;
  result := result mod max;
end;

{$IFDEF LIBRARY}
exports
  url_encode,
  find_codeset,
  split,
  octet_to_xdigit,
  byte_from_xdigit,
  swap2,
  swap4,
  swap8,
  swap16,
  string_hash;
{$ENDIF}
{$ELSE}
  function url_encode(val:  String): string; external MTDORB_Library_Name;
  function find_codeset(id:CodesetID): PCodesetInfo; external MTDORB_Library_Name;
  procedure split(const value: PChar; dest: TStrings); external MTDORB_Library_Name;
  function octet_to_xdigit(o: Octet): Octet; external MTDORB_Library_Name;
  function byte_from_xdigit(o: Octet): Byte; external MTDORB_Library_Name;
  procedure swap2(const src; var dst); external MTDORB_Library_Name;
  procedure swap4(const src; var dst); external MTDORB_Library_Name;
  procedure swap8(const src; var dst); external MTDORB_Library_Name;
  procedure swap16(const src; var dst); external MTDORB_Library_Name;
  function string_hash(str: string; max: _ulong): _ulong; external MTDORB_Library_Name;
implementation
{$ENDIF}
end.
